﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2017 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Problems.DataAnalysis;

namespace HeuristicLab.Algorithms.DataAnalysis {
  [StorableClass]
  internal class M5NodeModel : RegressionModel {
    #region Properties
    [Storable]
    internal bool IsLeaf { get; private set; }
    [Storable]
    internal IRegressionModel Model { get; set; }
    [Storable]
    internal string SplitAttribute { get; private set; }
    [Storable]
    internal double SplitValue { get; private set; }
    [Storable]
    internal M5NodeModel Left { get; private set; }
    [Storable]
    internal M5NodeModel Right { get; private set; }
    [Storable]
    internal M5NodeModel Parent { get; private set; }
    [Storable]
    internal int NumSamples { get; private set; }
    [Storable]
    private IReadOnlyList<string> variables;
    #endregion

    #region HLConstructors
    [StorableConstructor]
    protected M5NodeModel(bool deserializing) : base(deserializing) { }
    protected M5NodeModel(M5NodeModel original, Cloner cloner) : base(original, cloner) {
      IsLeaf = original.IsLeaf;
      Model = cloner.Clone(original.Model);
      SplitValue = original.SplitValue;
      SplitAttribute = original.SplitAttribute;
      Left = cloner.Clone(original.Left);
      Right = cloner.Clone(original.Right);
      Parent = cloner.Clone(original.Parent);
      NumSamples = original.NumSamples;
      variables = original.variables != null ? original.variables.ToList() : null;
    }
    private M5NodeModel(string targetAttr) : base(targetAttr) { }
    private M5NodeModel(M5NodeModel parent) : this(parent.TargetVariable) {
      Parent = parent;
    }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new M5NodeModel(this, cloner);
    }
    public static M5NodeModel CreateNode(string targetAttr, M5Parameters m5Params) {
      return m5Params.LeafModel.ProvidesConfidence ? new ConfidenceM5NodeModel(targetAttr) : new M5NodeModel(targetAttr);
    }
    private static M5NodeModel CreateNode(M5NodeModel parent, M5Parameters m5Params) {
      return m5Params.LeafModel.ProvidesConfidence ? new ConfidenceM5NodeModel(parent) : new M5NodeModel(parent);
    }
    #endregion

    #region RegressionModel
    public override IEnumerable<string> VariablesUsedForPrediction {
      get { return variables; }
    }
    public override IEnumerable<double> GetEstimatedValues(IDataset dataset, IEnumerable<int> rows) {
      if (!IsLeaf) return rows.Select(row => GetEstimatedValue(dataset, row));
      if (Model == null) throw new NotSupportedException("The model has not been built correctly");
      return Model.GetEstimatedValues(dataset, rows);
    }
    public override IRegressionSolution CreateRegressionSolution(IRegressionProblemData problemData) {
      return new RegressionSolution(this, problemData);
    }
    #endregion

    internal void Split(IReadOnlyList<int> rows, M5Parameters m5Params) {
      variables = m5Params.AllowedInputVariables.ToArray();
      NumSamples = rows.Count;
      Right = null;
      Left = null;
      Model = null;
      SplitAttribute = null;
      SplitValue = double.NaN;
      string attr;
      double splitValue;
      IsLeaf = !m5Params.Spliter.Split(new RegressionProblemData(M5StaticUtilities.ReduceDataset(m5Params.Data, rows, variables, TargetVariable), variables, TargetVariable), m5Params.MinLeafSize, out attr, out splitValue);
      if (IsLeaf) return;

      //split Dataset
      IReadOnlyList<int> leftRows, rightRows;
      M5StaticUtilities.SplitRows(rows, m5Params.Data, attr, splitValue, out leftRows, out rightRows);

      if (leftRows.Count < m5Params.MinLeafSize || rightRows.Count < m5Params.MinLeafSize) {
        IsLeaf = true;
        return;
      }
      SplitAttribute = attr;
      SplitValue = splitValue;

      //create subtrees
      Left = CreateNode(this, m5Params);
      Left.Split(leftRows, m5Params);
      Right = CreateNode(this, m5Params);
      Right.Split(rightRows, m5Params);
    }

    internal void ToLeaf() {
      IsLeaf = true;
      Right = null;
      Left = null;
    }

    internal void BuildLeafModels(IReadOnlyList<int> rows, M5Parameters parameters, CancellationToken cancellationToken) {
      if (!IsLeaf) {
        IReadOnlyList<int> leftRows, rightRows;
        M5StaticUtilities.SplitRows(rows, parameters.Data, SplitAttribute, SplitValue, out leftRows, out rightRows);
        Left.BuildLeafModels(leftRows, parameters, cancellationToken);
        Right.BuildLeafModels(rightRows, parameters, cancellationToken);
        return;
      }
      int numP;
      Model = M5StaticUtilities.BuildModel(rows, parameters, parameters.LeafModel, cancellationToken, out numP);
    }

    internal IEnumerable<M5NodeModel> EnumerateNodes() {
      var queue = new Queue<M5NodeModel>();
      queue.Enqueue(this);
      while (queue.Count != 0) {
        var cur = queue.Dequeue();
        yield return cur;
        if (cur.Left == null && cur.Right == null) continue;
        if (cur.Left != null) queue.Enqueue(cur.Left);
        if (cur.Right != null) queue.Enqueue(cur.Right);
      }
    }

    #region Helpers
    private double GetEstimatedValue(IDataset dataset, int row) {
      if (!IsLeaf) return (dataset.GetDoubleValue(SplitAttribute, row) <= SplitValue ? Left : Right).GetEstimatedValue(dataset, row);
      if (Model == null) throw new NotSupportedException("The model has not been built correctly");
      return Model.GetEstimatedValues(dataset, new[] {row}).First();
    }
    #endregion

    [StorableClass]
    private sealed class ConfidenceM5NodeModel : M5NodeModel, IConfidenceRegressionModel {
      #region HLConstructors
      [StorableConstructor]
      private ConfidenceM5NodeModel(bool deserializing) : base(deserializing) { }
      private ConfidenceM5NodeModel(ConfidenceM5NodeModel original, Cloner cloner) : base(original, cloner) { }
      public ConfidenceM5NodeModel(string targetAttr) : base(targetAttr) { }
      public ConfidenceM5NodeModel(M5NodeModel parent) : base(parent) { }
      public override IDeepCloneable Clone(Cloner cloner) {
        return new ConfidenceM5NodeModel(this, cloner);
      }
      #endregion

      public IEnumerable<double> GetEstimatedVariances(IDataset dataset, IEnumerable<int> rows) {
        return IsLeaf ? ((IConfidenceRegressionModel)Model).GetEstimatedVariances(dataset, rows) : rows.Select(row => GetEstimatedVariance(dataset, row));
      }

      private double GetEstimatedVariance(IDataset dataset, int row) {
        if (!IsLeaf)
          return ((IConfidenceRegressionModel)(dataset.GetDoubleValue(SplitAttribute, row) <= SplitValue ? Left : Right)).GetEstimatedVariances(dataset, row.ToEnumerable()).Single();
        return ((IConfidenceRegressionModel)Model).GetEstimatedVariances(dataset, new[] {row}).First();
      }

      public override IRegressionSolution CreateRegressionSolution(IRegressionProblemData problemData) {
        return new ConfidenceRegressionSolution(this, problemData);
      }
    }
  }
}